/* mgen.c
   microprogram generator
   from microprogram spec.  mspec.txt

   26 July 98
   modify for n-machine   27 Nov 2004
   modify for sx		  10 July 2006
*/

#include "mgen.h"

#define	 MXBUF  1024	   /* max length of input line */
#define	 MXMS	500		   /* max size of micro store */
#define	 MXW	128		   /* max width of micro store */

char 	ibuf[MXBUF], *cp = NULL;   /* input buffer, ch ptr */
static char sep[] = " \t\n";       /* separator char */
int 	lineno,loc;		   /* line num, current mpgm ads */
int 	width = 0, awidth;         /* width of mpgm */
int 	pass;
FILE 	*fi, *fo, *fo2;

char 	ms[MXMS][MXW];		   /* micro store */

void seterr(char *s){
  printf("line %d error : %s\n",lineno,s);
  exit(0);
}

int eq( char *s1, char *s2){	/* string equal */
  return( strcmp(s1,s2) == 0);
}

char *tok(void){	/* get one token from input */
  if( cp != NULL )
	  cp = strtok(NULL,sep);
  while ( cp == NULL ) {
    if( fgets(ibuf,MXBUF,fi) == NULL ) return NULL;
    lineno++;
	if(pass == 1)
		fprintf(fo2,"%d %s",loc,ibuf);
    cp = strtok(ibuf, sep);
	if( eq(cp,"..") ) cp = NULL;  /* comment, skip */
  }
//  printf("%s ",cp);
  return cp;
}

int iscmd( char *w ){	/* Is token a command ? */
  return ( w[0] == '.' );
}

static char ww[80];

// substitute dot with underscore
char *dedot(char *s){
	char c, *s2;
	s2 = ww;
	while( (c = *s) != 0 ){
		if( c == '.' ) c = '_';
		*s2 = c;
		s++;
		s2++;
	}
	*s2 = 0;
	return ww;
}

/* get token update symbol table until other cmd */
void dosymbol(void){
  char *w;
  int v;
  w = tok();
  while ( w != NULL && ! iscmd(w) ) {
	putsym(w,width);
	fprintf(fo,"#define s_%s %d\n",dedot(w),width);
    width++;
	w = tok();
  }
}
/* get label update symbol table until other cmd */
void dolabel(void){
  char *w;
  double x;
  w = tok();
  while ( w != NULL && ! iscmd(w) ) {
    if( w[0] == ':' ){
	   putsym(w+1,loc);
	   fprintf(fo,"#define a_%s %d\n",dedot(w+1),loc);
	}
    if( eq(w,";") ) loc++;
    w = tok();
  }
  x = log2((double) loc);
  awidth = (int) x;
  if( x > (double)awidth ) awidth += 1;
  fprintf(fo,"#define MCWIDTH %d\n",width);
  fprintf(fo,"#define MAWIDTH %d\n",awidth);
  fprintf(fo,"#define MLEN %d\n",loc);
}

void pass1(void){		/* collect symbols */
 char *w;
 pass = 1;
 w = tok();
 while ( iscmd(w) ) {
   switch( w[1] ) {
   case 'e' : return;
   case 's' : dosymbol(); break;
   case 'm' : dolabel(); break;
   }
 }
}

void putads(int a){
  int i,a2;
  a2 = width + awidth;
  for( i=a2-1; i>=width; i-- ){
    ms[loc][i] = a & 1 ? '1' : '0';
	a = a >> 1;
  }
}

void pass2(void){		/* generate microprogram */
  char *w;
  int a, flag;
  pass = 2;
  w = tok();
  while ( w != NULL ) {		/* skip to .m section */
    if( eq(w,".e")) return;
    if( eq(w,".m")) break;
    w = tok();
  }
  w = tok();
  flag = 0;
  while ( w != NULL ) {
    if( eq(w,".e")) return;
    switch( w[0] ) {
    case ':' : break;
    case ';' :
		if (flag == 0)
			putads(loc+1);  // default next ads
		flag = 0;
		loc++;
		break;
    case '/' :
      a = getsym(w+1);
      if ( a < 0 ){
		 printf("%s ",w);
		 seterr("symbol not found");
	  }
      putads(a);
	  flag = 1;
      break;
    default :
      a = getsym(w);
      if ( a < 0 ){
		 printf("%s ",w);
		 seterr("symbol not found");
	  }
      ms[loc][a] = '1';
      break;
    }
    w = tok();
  }
}

void printout(int n){
  int i,j;
  printf("%d\n",n);
  for( i=0; i<n; i++ ) {
    printf("%3d ",i);
    for( j=0; j<width+awidth; j++ )
      printf("%c",ms[i][j]);
    printf("\n");
  }
}

void initms(int height, int width){
  int i,j;
  for( i=0; i<height; i++ )
    for( j=0; j<width; j++ )
      ms[i][j] = '0';
}

int main(void){
  fi = fopen("mspec.txt","r");
  fo = fopen("mspec.h","w");
  fo2 = fopen("mlist.txt","w");
  lineno = 0;
  loc = 0;
  pass1();
  fclose(fo2);
  fclose(fo);
  rewind(fi);
  initms(loc,width+awidth);
  lineno = 0;
  loc = 0;
  pass2();
  fclose(fi);
  printout(loc);
  return 0;
}

